@@ -35,6 +35,9 @@ typedef struct {
 	int flags;
 	git_atomic cancelled;
 	git_repository *repo;
+	git_transport_message_cb progress_cb;
+	git_transport_message_cb error_cb;
+	void *message_cb_payload;
 	git_vector refs;
 	unsigned connected : 1,
 		have_refs : 1;
@@ -47,6 +50,17 @@ static void free_head(git_remote_head *head)
 	git__free(head);
 }
 
+static void free_heads(git_vector *heads)
+{
+	git_remote_head *head;
+	size_t i;
+
+	git_vector_foreach(heads, i, head)
+		free_head(head);
+
+	git_vector_free(heads);
+}
+
 static int add_ref(transport_local *t, const char *name)
 {
 	const char peeled[] = "^{}";
@@ -115,8 +129,10 @@ static int add_ref(transport_local *t, const char *name)
 	head = git__calloc(1, sizeof(git_remote_head));
 	GITERR_CHECK_ALLOC(head);
 
-	if (git_buf_join(&buf, 0, name, peeled) < 0)
+	if (git_buf_join(&buf, 0, name, peeled) < 0) {
+		free_head(head);
 		return -1;
+	}
 	head->name = git_buf_detach(&buf);
 
 	if (!(error = git_tag_peel(&target, (git_tag *)obj))) {
@@ -198,6 +214,8 @@ static int local_connect(
 	if (t->connected)
 		return 0;
 
+	free_heads(&t->refs);
+
 	t->url = git__strdup(url);
 	GITERR_CHECK_ALLOC(t->url);
 	t->direction = direction;
@@ -326,14 +344,11 @@ static int local_push_update_remote_ref(
 	int error;
 	git_reference *remote_ref = NULL;
 
-	/* rref will be NULL if it is implicit in the pushspec (e.g. 'b1:') */
-	rref = rref ? rref : lref;
-
-	if (lref) {
+	/* check for lhs, if it's empty it means to delete */
+	if (lref[0] != '\0') {
 		/* Create or update a ref */
-		if ((error = git_reference_create(NULL, remote_repo, rref, loid,
-				!git_oid_iszero(roid), NULL, NULL)) < 0)
-			return error;
+		error = git_reference_create(NULL, remote_repo, rref, loid,
+					     !git_oid_iszero(roid), NULL);
 	} else {
 		/* Delete a ref */
 		if ((error = git_reference_lookup(&remote_ref, remote_repo, rref)) < 0) {
@@ -342,13 +357,11 @@ static int local_push_update_remote_ref(
 			return error;
 		}
 
-		if ((error = git_reference_delete(remote_ref)) < 0)
-			return error;
-
+		error = git_reference_delete(remote_ref);
 		git_reference_free(remote_ref);
 	}
 
-	return 0;
+	return error;
 }
 
 static int local_push(
@@ -405,9 +418,9 @@ static int local_push(
 	git_vector_foreach(&push->specs, j, spec) {
 		push_status *status;
 		const git_error *last;
-		char *ref = spec->rref ? spec->rref : spec->lref;
+		char *ref = spec->refspec.dst;
 
-		status = git__calloc(sizeof(push_status), 1);
+		status = git__calloc(1, sizeof(push_status));
 		if (!status)
 			goto on_error;
 
@@ -417,7 +430,7 @@ static int local_push(
 			goto on_error;
 		}
 
-		error = local_push_update_remote_ref(remote_repo, spec->lref, spec->rref,
+		error = local_push_update_remote_ref(remote_repo, spec->refspec.src, spec->refspec.dst,
 			&spec->loid, &spec->roid);
 
 		switch (error) {
@@ -486,6 +499,8 @@ static int foreach_cb(void *buf, size_t len, void *payload)
 	return data->writepack->append(data->writepack, buf, len, data->stats);
 }
 
+static const char *counting_objects_fmt = "Counting objects %d\r";
+
 static int local_download_pack(
 		git_transport *transport,
 		git_repository *repo,
@@ -498,10 +513,10 @@ static int local_download_pack(
 	git_remote_head *rhead;
 	unsigned int i;
 	int error = -1;
-	git_oid oid;
 	git_packbuilder *pack = NULL;
 	git_odb_writepack *writepack = NULL;
 	git_odb *odb = NULL;
+	git_buf progress_info = GIT_BUF_INIT;
 
 	if ((error = git_revwalk_new(&walk, t->repo)) < 0)
 		goto cleanup;
@@ -523,38 +538,42 @@ static int local_download_pack(
 		if (git_object_type(obj) == GIT_OBJ_COMMIT) {
 			/* Revwalker includes only wanted commits */
 			error = git_revwalk_push(walk, &rhead->oid);
-			if (!git_oid_iszero(&rhead->loid))
+			if (!error && !git_oid_iszero(&rhead->loid)) {
 				error = git_revwalk_hide(walk, &rhead->loid);
+				if (error == GIT_ENOTFOUND)
+					error = 0;
+			}
 		} else {
-			/* Tag or some other wanted object. Add it on its own */
 			error = git_packbuilder_insert(pack, &rhead->oid, rhead->name);
 		}
 		git_object_free(obj);
+		if (error < 0)
+			goto cleanup;
 	}
 
+	if ((error = git_packbuilder_insert_walk(pack, walk)))
+		goto cleanup;
+
+	if ((error = git_buf_printf(&progress_info, counting_objects_fmt, git_packbuilder_object_count(pack))) < 0)
+		goto cleanup;
+
+	if (t->progress_cb &&
+	    (error = t->progress_cb(git_buf_cstr(&progress_info), git_buf_len(&progress_info), t->message_cb_payload)) < 0)
+		goto cleanup;
+
 	/* Walk the objects, building a packfile */
 	if ((error = git_repository_odb__weakptr(&odb, repo)) < 0)
 		goto cleanup;
 
-	while ((error = git_revwalk_next(&oid, walk)) == 0) {
-		git_commit *commit;
+	/* One last one with the newline */
+	git_buf_clear(&progress_info);
+	git_buf_printf(&progress_info, counting_objects_fmt, git_packbuilder_object_count(pack));
+	if ((error = git_buf_putc(&progress_info, '\n')) < 0)
+		goto cleanup;
 
-		/* Skip commits we already have */
-		if (git_odb_exists(odb, &oid)) continue;
-
-		if (!git_object_lookup((git_object**)&commit, t->repo, &oid, GIT_OBJ_COMMIT)) {
-			const git_oid *tree_oid = git_commit_tree_id(commit);
-
-			/* Add the commit and its tree */
-			if ((error = git_packbuilder_insert(pack, &oid, NULL)) < 0 ||
-				 (error = git_packbuilder_insert_tree(pack, tree_oid)) < 0) {
-				git_commit_free(commit);
-				goto cleanup;
-			}
-
-			git_commit_free(commit);
-		}
-	}
+	if (t->progress_cb &&
+	    (error = t->progress_cb(git_buf_cstr(&progress_info), git_buf_len(&progress_info), t->message_cb_payload)) < 0)
+		goto cleanup;
 
 	if ((error = git_odb_write_pack(&writepack, odb, progress_cb, progress_payload)) != 0)
 		goto cleanup;
@@ -567,18 +586,41 @@ static int local_download_pack(
 		data.progress_payload = progress_payload;
 		data.writepack = writepack;
 
+		/* autodetect */
+		git_packbuilder_set_threads(pack, 0);
+
 		if ((error = git_packbuilder_foreach(pack, foreach_cb, &data)) != 0)
 			goto cleanup;
 	}
+
 	error = writepack->commit(writepack, stats);
 
 cleanup:
 	if (writepack) writepack->free(writepack);
+	git_buf_free(&progress_info);
 	git_packbuilder_free(pack);
 	git_revwalk_free(walk);
 	return error;
 }
 
+static int local_set_callbacks(
+	git_transport *transport,
+	git_transport_message_cb progress_cb,
+	git_transport_message_cb error_cb,
+	git_transport_certificate_check_cb certificate_check_cb,
+	void *message_cb_payload)
+{
+	transport_local *t = (transport_local *)transport;
+
+	GIT_UNUSED(certificate_check_cb);
+
+	t->progress_cb = progress_cb;
+	t->error_cb = error_cb;
+	t->message_cb_payload = message_cb_payload;
+
+	return 0;
+}
+
 static int local_is_connected(git_transport *transport)
 {
 	transport_local *t = (transport_local *)transport;
@@ -624,13 +666,8 @@ static int local_close(git_transport *transport)
 static void local_free(git_transport *transport)
 {
 	transport_local *t = (transport_local *)transport;
-	size_t i;
-	git_remote_head *head;
 
-	git_vector_foreach(&t->refs, i, head)
-		free_head(head);
-
-	git_vector_free(&t->refs);
+	free_heads(&t->refs);
 
 	/* Close the transport, if it's still open. */
 	local_close(transport);
@@ -645,6 +682,7 @@ static void local_free(git_transport *transport)
 
 int git_transport_local(git_transport **out, git_remote *owner, void *param)
 {
+	int error;
 	transport_local *t;
 
 	GIT_UNUSED(param);
@@ -653,6 +691,7 @@ int git_transport_local(git_transport **out, git_remote *owner, void *param)
 	GITERR_CHECK_ALLOC(t);
 
 	t->parent.version = GIT_TRANSPORT_VERSION;
+	t->parent.set_callbacks = local_set_callbacks;
 	t->parent.connect = local_connect;
 	t->parent.negotiate_fetch = local_negotiate_fetch;
 	t->parent.download_pack = local_download_pack;
@@ -664,7 +703,11 @@ int git_transport_local(git_transport **out, git_remote *owner, void *param)
 	t->parent.read_flags = local_read_flags;
 	t->parent.cancel = local_cancel;
 
-	git_vector_init(&t->refs, 0, NULL);
+	if ((error = git_vector_init(&t->refs, 0, NULL)) < 0) {
+		git__free(t);
+		return error;
+	}
+
 	t->owner = owner;
 
 	*out = (git_transport *) t;
